/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of the
 * License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */

#include "stdafx.h"

#include "starter_popup.h"
#include "mforms/utilities.h"
#include "string_utilities.h"

using namespace wb;
using namespace mforms;

#ifdef __APPLE__
#define STARTER_TITLE_FONT "Lucida Grande"
#define STARTER_DESCRIPTION_FONT "Lucida Grande"
#elif _WIN32
#define STARTER_TITLE_FONT "Tahoma"
#define STARTER_DESCRIPTION_FONT "Tahoma"
#else
#define STARTER_TITLE_FONT "Helvetica"
#define STARTER_DESCRIPTION_FONT "Helvetica"
#endif

#define STARTER_TITLE_FONT_SIZE 13
#define STARTER_DESCRIPTION_FONT_SIZE 11
#define STARTER_BUTTON_FONT_SIZE 13

//--------------------------------------------------------------------------------------------------

// The following helpers are just temporary. They will be replaced by a cairo context class.
static void delete_surface(cairo_surface_t* surface)
{
  if (surface != NULL)
    cairo_surface_destroy(surface);
}

//--------------------------------------------------------------------------------------------------

static void delete_pattern(cairo_pattern_t* pattern)
{
  if (pattern != NULL)
    cairo_pattern_destroy(pattern);
}

//--------------------------------------------------------------------------------------------------

static int image_width(cairo_surface_t* image)
{
  if (image != NULL)
    return cairo_image_surface_get_width(image);
  return 0;
}

//--------------------------------------------------------------------------------------------------

static int image_height(cairo_surface_t* image)
{
  if (image != NULL)
    return cairo_image_surface_get_height(image);
  return 0;
}

//----------------- StarterPopup -------------------------------------------------------------------

StarterPopup::StarterPopup(WBContext* context)
: Popup(mforms::PopupBezel), _context(context)
{
  _mysql_plugin_overlay= Utilities::load_icon("wb_starter_type_mysql.png");
  _community_plugin_overlay= Utilities::load_icon("wb_starter_type_community.png");
  _legal_image= Utilities::load_icon("wb_starter_legal.png");

  _button_parts[0]= Utilities::load_icon("wb_starter_button_left.png");
  _button_parts[1]= Utilities::load_icon("wb_starter_button_middle.png");
  _button_parts[2]= Utilities::load_icon("wb_starter_button_right.png");

  if (_button_parts[1] != NULL)
  {
    _button_pattern = cairo_pattern_create_for_surface(_button_parts[1]);
    cairo_pattern_set_extend(_button_pattern, CAIRO_EXTEND_REPEAT);
  }
  else
    _button_pattern= NULL;}

//--------------------------------------------------------------------------------------------------

StarterPopup::~StarterPopup()
{
  delete_surface(_mysql_plugin_overlay);
  delete_surface(_community_plugin_overlay);
  delete_surface(_legal_image);

  delete_surface(_button_parts[0]);
  delete_surface(_button_parts[1]);
  delete_surface(_button_parts[2]);
  if (_button_pattern != NULL)
    delete_pattern(_button_pattern);}

//--------------------------------------------------------------------------------------------------

/**
 * Draws a starter button.
 */
void StarterPopup::draw_button(cairo_t* cr, double x, double y, double inner_width, double h)
{
  if (_button_pattern != NULL)
  {
    cairo_set_source_surface(cr, _button_parts[0], x, y);
    cairo_paint(cr);
    
    x += image_width(_button_parts[0]);
    
    // In order to align the pattern properly to our target bounds we create a
    // translation transformation with the desired amount.
    cairo_matrix_t matrix;
    cairo_matrix_init_translate(&matrix, -x, -y);
    cairo_pattern_set_matrix(_button_pattern, &matrix);
    
    cairo_set_source(cr, _button_pattern);
    cairo_move_to(cr, x, y);
    cairo_rectangle(cr, x, y, inner_width, h);
    cairo_fill(cr);
    
    cairo_set_source_surface(cr, _button_parts[2], x + inner_width, y);
    cairo_paint(cr);
  }
}

//--------------------------------------------------------------------------------------------------

#define STARTER_OVERLAY_OFFSET 8    // Vertical offset for the entry to make the overlay icon stand out a bit
                                    // and start at top position 0.
#define STARTER_TEXT_SPACING 12     // Horizontal distance between icon and text area.
#define STARTER_INFO_TEXT_SPACING 6 // Horizontal distance between info text parts.
#define STARTER_SPACING 20          // Spacing between adjacent starter entries.
#define STARTER_TEXT_LINE_HEIGHT 14 // Hard coded line height, to make the overall image more compact.
#define STARTER_BUTTON_OFFSET 225   // Vertical offset for the buttons.

#define STARTER_PUBLISHER_STRING _("Publisher:")
#define STARTER_TYPE_STRING _("Type:")

void StarterPopup::draw_entry(cairo_t* cr, app_StarterRef starter, int x, int y, int w, int h)
{
  cairo_surface_t* icon= Utilities::load_icon(starter->largeIcon());
  cairo_set_source_surface(cr, icon, x, y + STARTER_OVERLAY_OFFSET);
  cairo_paint(cr);
  int icon_width= image_width(icon);
  delete_surface(icon);

  if (starter->publisher() == "Oracle Corp.")
    icon= _mysql_plugin_overlay;
  else
    icon= _community_plugin_overlay;
  int overlay_width= image_width(icon);
  cairo_set_source_surface(cr, icon, x + icon_width - overlay_width + 4, y);
  cairo_paint(cr);

  double space_for_text= w - icon_width - STARTER_TEXT_SPACING - STARTER_SPACING;

  // Title.
  cairo_select_font_face(cr, STARTER_TITLE_FONT, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cr, STARTER_TITLE_FONT_SIZE);
  double text_offset= x + icon_width + STARTER_TEXT_SPACING;
  cairo_set_source_rgb(cr, 1, 1, 1);

  // Usually lines are set with the vertical distance in extents.height, however since have not much space
  // we only use the minimal height possible.
  double line1_offset= y + STARTER_OVERLAY_OFFSET + STARTER_TEXT_LINE_HEIGHT / 2;
  cairo_move_to(cr, text_offset, line1_offset);

  // We add a space char here as cairo sometimes not not fully render the last character.
  // Better this wrong "glyph" is a space char.
  std::string text= Utilities::shorten_string(cr, *(starter->title()) + " ", space_for_text);
  cairo_show_text(cr, text.c_str());
  cairo_stroke(cr);

  // Description. We display at most two lines of the description. If it is longer then we show end ellipses.
  // TODO: once Pango is available use its word wrapping abilities.
  cairo_set_source_rgba(cr, 170 / 255.0, 173 / 255.0, 181 / 255.0, 1);
  cairo_select_font_face(cr, STARTER_DESCRIPTION_FONT, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cr, STARTER_DESCRIPTION_FONT_SIZE);

  line1_offset += STARTER_TEXT_LINE_HEIGHT;
  double line2_offset = line1_offset + STARTER_TEXT_LINE_HEIGHT;
  std::string description= starter->description();
  gchar* head= (gchar*) description.c_str();
  const gchar* end;
  g_utf8_validate(head, -1, &end); // Display only the valid part of the string.
  gchar* tail= head;

  std::string line1;
  std::string line2;
  while (true)
  {
    // Get the next word.
    while (tail != end && (*tail != ' '))
      tail= g_utf8_next_char(tail);
    gchar* part= g_strndup(head, tail - head);

    // See which space the part requires.
    cairo_text_extents_t local_extents;
    cairo_text_extents(cr, part, &local_extents);
    if ((tail == end) || ceil(local_extents.width) >= space_for_text)
    {
      g_free(part);
      break;
    }

    line1= part;
    g_free(part);

    // Skip over the space char.
    tail= g_utf8_next_char(tail);
  }
  cairo_move_to(cr, text_offset, line1_offset);
  cairo_show_text(cr, line1.c_str());
  cairo_stroke(cr);

  // Get the rest of the string and determine space requirement.
  // Skip over the split space char if we found one.
  if (tail != end)
    tail= g_utf8_next_char(tail);
  gchar* part= g_strndup(tail, end - tail);
  line2= Utilities::shorten_string(cr, part, space_for_text);
  cairo_move_to(cr, text_offset, line2_offset);
  cairo_show_text(cr, line2.c_str());
  cairo_stroke(cr);

  // Additional info. Lines start with right aligned text.
  line1_offset += 2 * STARTER_TEXT_LINE_HEIGHT; // For publisher line.
  line2_offset += 2 * STARTER_TEXT_LINE_HEIGHT; // For type line.
  double info_offset;
  cairo_text_extents_t text_extents;
  cairo_text_extents(cr, STARTER_PUBLISHER_STRING, &text_extents);
  double part1_width= text_extents.width;
  info_offset= text_extents.width;
  cairo_text_extents(cr, STARTER_TYPE_STRING, &text_extents);
  double part2_width= text_extents.width;
  if (text_extents.width > info_offset)
    info_offset= text_extents.width;
  space_for_text -= info_offset;
  info_offset += text_offset;

  cairo_move_to(cr, info_offset - part1_width, line1_offset);
  cairo_show_text(cr, STARTER_PUBLISHER_STRING);
  cairo_stroke(cr);

  cairo_move_to(cr, info_offset - part2_width, line2_offset);
  cairo_show_text(cr, STARTER_TYPE_STRING);
  cairo_stroke(cr);

  cairo_set_source_rgb(cr, 1, 1, 1);

  text= Utilities::shorten_string(cr, *(starter->publisher()), space_for_text);
  cairo_move_to(cr, info_offset + STARTER_INFO_TEXT_SPACING, line1_offset);
  cairo_show_text(cr, text.c_str());
  cairo_stroke(cr);

  text= Utilities::shorten_string(cr, *(starter->type()), space_for_text);
  cairo_move_to(cr, info_offset + STARTER_INFO_TEXT_SPACING, line2_offset);
  cairo_show_text(cr, text.c_str());
  cairo_stroke(cr);
}

//--------------------------------------------------------------------------------------------------

#define STARTER_BUTTON_PADDING 7        // Space left and right of the button text
#define STARTER_MORE_PLUGINS_BUTTON_MIN_WIDTH 140 // Minimal width of the "Get More Plugins.." button.
#define STARTER_NAV_BUTTON_MIN_WIDTH 80 // Minimal width of the navigation buttons.
#define STARTER_BUTTON_HEIGHT 32        // Height of the buttons.
#define STARTER_BUTTON_SPACING 8        // Spacing between back and next button.

#define STARTER_MORE_PLUGINS_TEXT _("Get More Plugins...")
#define STARTER_BACK_TEXT _("< Back")
#define STARTER_NEXT_TEXT _("Next >")

/**
 * This is the core drawing routine for the popup content.
 * Note: we take advantage of the knowledge that we draw on a background bitmap, so nothing is
 * buffered here, except for structures needed for hit testing.
 */
void StarterPopup::repaint(cairo_t *cr, int x, int y, int w, int h)
{
  // We want text anti-aliasing with gray pixels, as we render white text on a black background.
  cairo_font_options_t* options= cairo_font_options_create();
  cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_GRAY);

  // TODO: the gray setting doesn't seem to be supported everywhere. At least on my Win7 VM
  //       it has the effect to render without any anti-aliasing. Hence it is disabled for now.
  //       Might later be enabled again when it works as expected on all platforms.
  //cairo_set_font_options(cr, options);

  int current_left= x;
  int current_top= y;

  // 6 entries per page. Start with the predefined starters.
  int counter= 6;
  grt::ListRef<app_Starter> starters= _context->get_root()->starters()->predefined();
  for (grt::ListRef<app_Starter>::const_iterator iterator= starters.begin();
    iterator != starters.end() && counter > 0;
    iterator++, counter--)
  {
    draw_entry(cr, *iterator, current_left, current_top, 270, 75);

    current_left += (int) ceil(w / 3.0);
    if (current_left >= x + w)
    {
      // Advance to next line.
      current_left= x;
      current_top += 110;
    }
  }

  // Now the user starters.
  starters= _context->get_root()->starters()->custom();
  for (grt::ListRef<app_Starter>::const_iterator iterator= starters.begin();
    iterator != starters.end() && counter > 0;
    iterator++, counter--)
  {
    draw_entry(cr, *iterator, current_left, current_top, 270 , 75);

    current_left += (int) ceil(w / 3.0);
    if (current_left >= x + w)
    {
      // Advance to next line.
      current_left= x;
      current_top += 110;
    }
  }

  // Buttons.
  cairo_select_font_face(cr, STARTER_TITLE_FONT, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cr, STARTER_BUTTON_FONT_SIZE);

  double button_top = y + STARTER_BUTTON_OFFSET; // Same for all buttons.
  double button_left= x;

  cairo_text_extents_t extents;
  cairo_text_extents(cr, STARTER_MORE_PLUGINS_TEXT, &extents);
  
  // Text position, center vertically.
  current_left= x + image_width(_button_parts[0]) + STARTER_BUTTON_PADDING;
  current_top= (int) ceil(button_top + (STARTER_BUTTON_HEIGHT - extents.height) / 2 - extents.y_bearing);

  double button_width= ceil(extents.width) + 2 * STARTER_BUTTON_PADDING;
  double outer_offset= image_width(_button_parts[0]) + image_width(_button_parts[2]);
  if (button_width < STARTER_MORE_PLUGINS_BUTTON_MIN_WIDTH - outer_offset)
    button_width= STARTER_MORE_PLUGINS_BUTTON_MIN_WIDTH - outer_offset;
  draw_button(cr, button_left, button_top, button_width, STARTER_BUTTON_HEIGHT);
  _more_button_bounds.width= button_width + outer_offset;
  _more_button_bounds.height= STARTER_BUTTON_HEIGHT;
  _more_button_bounds.left= button_left;
  _more_button_bounds.top= button_top;

  cairo_move_to(cr, current_left, current_top);
  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_show_text(cr, STARTER_MORE_PLUGINS_TEXT);
  cairo_stroke(cr);

  // Next and back buttons are right aligned, so we need their width first. Both buttons will be set to
  // have the same width, depending on which is larger.
  cairo_text_extents(cr, STARTER_BACK_TEXT, &extents);
  button_width= ceil(extents.width);
  double back_width= button_width; // For horizontal text alignment.
  cairo_text_extents(cr, STARTER_NEXT_TEXT, &extents);
  double next_width= ceil(extents.width); // For horizontal text alignment.
  if (extents.width > next_width)
    button_width= next_width;
  button_width += 2 * STARTER_BUTTON_PADDING;
  if (button_width < STARTER_NAV_BUTTON_MIN_WIDTH - outer_offset)
    button_width= STARTER_NAV_BUTTON_MIN_WIDTH - outer_offset;
  button_left = x + w - 2 * (button_width + outer_offset) - STARTER_BUTTON_SPACING;
  draw_button(cr, button_left, button_top, button_width, STARTER_BUTTON_HEIGHT);
  _back_button_bounds.left= button_left;
  _back_button_bounds.top= button_top;
  _back_button_bounds.width= button_width + outer_offset;
  _back_button_bounds.height= STARTER_BUTTON_HEIGHT;

  current_left= (int) ceil(button_left + (_back_button_bounds.width - back_width) / 2);
  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_move_to(cr, current_left, current_top);
  cairo_show_text(cr, STARTER_BACK_TEXT);
  cairo_stroke(cr);

  button_left += _back_button_bounds.width + STARTER_BUTTON_SPACING;
  _next_button_bounds.left= button_left;
  _next_button_bounds.top= button_top;
  _next_button_bounds.width= button_width + outer_offset;
  _next_button_bounds.height= STARTER_BUTTON_HEIGHT;
  current_left= (int) ceil(button_left + (_next_button_bounds.width - next_width) / 2);

  draw_button(cr, button_left, button_top, button_width, STARTER_BUTTON_HEIGHT);

  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_move_to(cr, current_left, current_top);
  cairo_show_text(cr, STARTER_NEXT_TEXT);
  cairo_stroke(cr);

  // Legal text.
  cairo_set_source_surface(cr, _legal_image, x, y + h - image_height(_legal_image));
  cairo_paint(cr);

  // Clean up.
  cairo_font_options_destroy(options);
}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_down(int button, int x, int y)
{

}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_up(int button, int x, int y)
{

}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_click(int button, int x, int y)
{
}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_enter()
{
}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_leave()
{
}

//--------------------------------------------------------------------------------------------------

void StarterPopup::mouse_move(int x, int y)
{
}

//--------------------------------------------------------------------------------------------------

